#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;
use iANT::Cfg;
use iANT::Common;
use iANT::DB::Entry;
require "iANT/Utils.pl";
use MIME::Base64;
use File::Temp;
use Cwd 'abs_path';
use File::Basename;
use General;

my $SLICE_LUCENE_QUERY = 50;



MAIN:
{

    my $o_param = New ParamParser(
                                   "GETOPTLONG",\&Usage,
                                   ("parser_cfg=s","store_dir=s" ,"outdir=s","output_field=s","light_entry_xml","outfile_suffix=s", 'table', 'faa', 'verbose')
                                     );


    $o_param->AssertDefined('outdir');
    $o_param->AssertDirExists('store_dir');
    $o_param->AssertFileExists('parser_cfg');
    
    $o_param->SetBehaviour('use_substitution_table');
    $o_param = &iANT::Cfg::Load($o_param);
    if ($o_param->IsDefined('table'))
    {
		&Idx2Xls($o_param);
	}
	elsif($o_param->IsDefined('faa'))
	{
		&Idx2Faa($o_param);	
	}
	else
	{
	    my $outdir = $o_param->Get('outdir');
	    mkdir $outdir if (! -d $outdir);
	
	    
	    
	    my $parser = $o_param->Get('parser_cfg');
	    my $store = $o_param->Get('store_dir');
	    
	    
	    my $o_parser     = New ParamParser($parser);
	    my $docid_field = $o_parser->Get('docid');
	    $o_param->SetUnlessDefined('output_field', $o_parser->Get('default_out'));
	
	    $o_parser->SetUnlessDefined('suffix',$docid_field);
	    my $suffix = $o_parser->Get('suffix');
	    $suffix =~ s/^\.//;
	    
	    my $output_field = $o_param->Get('output_field');
	    
	    
	    my $lucene_searcher = $o_param->Get('lucene_searcher');
	    my @a_docids = `$lucene_searcher --parser_cfg=$parser --store_dir=$store --list --query=$docid_field`;
	    chomp @a_docids;
	
	    my $fasta2entrydb_max_tree_subdirectories = $o_param->Get('fasta2entrydb_max_tree_subdirectories');
	    
	    my $subdir = 1;
	    $subdir = sprintf ("%04d", $subdir);
	    my $count = 0;
	
	
	    foreach my $docid_value (@a_docids)
	    {
	            next if ($docid_value eq '');        
	            if ($count == $fasta2entrydb_max_tree_subdirectories)
	            {
	                $subdir ++;
	                $subdir = sprintf ("%04d", $subdir);
	                $count = 0;
	            }
	            $count ++;
	            
	
	            mkdir "$outdir/$subdir" if (! -d "$outdir/$subdir");
	            mkdir "$outdir/$subdir/$docid_value" if (! -d "$outdir/$subdir/$docid_value");
	            
	            my $cleaning_cmd = '';
	            if ($o_param->IsDefined('light_entry_xml'))
	            {
	                $cleaning_cmd = '|perl -ne \'$_ =~ s/\n/__nl__/g;print\' | perl -ne \'$_ =~ s/<entry_features.+<\/entry_features>/<entry_features\/>/g;$_ =~ s/(\s*__nl__\s*)+/\n/g;print\' | xmllint --format -';
	            }
	            `$lucene_searcher  --parser_cfg=$parser --store_dir=$store --query=$docid_field:$docid_value --output_fields=$output_field --no_output_field_name $cleaning_cmd > $outdir/$subdir/$docid_value/$docid_value.$suffix`;
	    }
	}
}

sub Idx2Faa
{

    my ($o_param) = @_;
    
    my $parser_cfg = $o_param->Get('parser_cfg');
    my $store_dir = $o_param->Get('store_dir');
	my $lucene_searcher = $o_param->Get('lucene_searcher');
	
    my $o_param_idx = New ParamParser($parser_cfg);
	
	my $docid_field = $o_param_idx->Get('docid');
    my $lucene_searcher = $o_param->Get('lucene_searcher');
    my @a_docids = `$lucene_searcher --parser_cfg=$parser_cfg --store_dir=$store_dir --list --query=$docid_field`;
    
	
	chomp @a_docids;
    
    my $fasta = '';
    
    foreach my $doc_id (@a_docids)
    {
		
		my $lucene_query = "$docid_field:$doc_id";

    	my $cleaning_cmd = '';
	    my $cleaning_cmd = '|perl -ne \'$_ =~ s/\n/__nl__/g;print\' | perl -ne \'$_ =~ s/<entry_features.+<\/entry_features>/<entry_features\/>/g;$_ =~ s/(\s*__nl__\s*)+/\n/g;print\' | xmllint --format -';
	    
		
		my $cde_get_definitions = "$lucene_searcher -query '$lucene_query' -store $store_dir -parser $parser_cfg  --no_output_field_name $cleaning_cmd ";
        print STDERR "CMD - $cde_get_definitions\n" if ($o_param->IsDefined('verbose'));
        my $raw_results = `$cde_get_definitions`;
        
        my $o_db_entries = New iANT::DB::Entry($raw_results);
        my @a_o_entry = @{$o_db_entries->GetEntries};
        foreach my $o_entry (@a_o_entry)
        {
				print $o_entry->GetLocus_tag . "\n";
				if ($o_entry->GetType eq 'protein')
				{
						$fasta .= $o_entry->GetFasta;
				}
				else
				{
						print $o_entry->GetLocus_tag . ' ' . $o_entry->GetType . "\n";
				}
		}
		 
	}
	
	my $outdir = $o_param->Get('outdir');
    mkdir $outdir unless (-d $outdir);
    my $date = `date +%Y%m%d`;
    chomp $date;
    my $type = basename($store_dir);
    my $outfile = "$outdir/$date.$type.faa";
    my $fh_out = &GetStreamOut($outfile);

    print STDERR "INFO - Print output file $outfile\n" if ($o_param->IsDefined('verbose'));
  	print $fh_out $fasta;
  	$fh_out->close;
    
    return;

}

sub Idx2Xls
{

    my ($o_param) = @_;
    my %h_dbnames = ();
    my %h_analyses = ();
	my %h_deflines = ();
    my %h_absolutebesthit = ();
    
    my $parser_cfg = $o_param->Get('parser_cfg');
    my $store_dir = $o_param->Get('store_dir');
	my $lucene_searcher = $o_param->Get('lucene_searcher');
	
    my $o_param_idx = New ParamParser($parser_cfg);
    my @a_dbnames = split (',', $o_param_idx->Get('hit_bank_complete'));
    
    if (scalar  @a_dbnames  == 0)
    {
			Carp::croak ("ERROR - NO bank found in $parser_cfg (hit_bank_complete value is empty)");
	}
	
	print STDERR "INFO - Report best hits for banks : " . join ("\t", @a_dbnames) . "\n" if ($o_param->IsDefined('verbose'));
	
	my $docid_field = $o_param_idx->Get('docid');
    my $lucene_searcher = $o_param->Get('lucene_searcher');
    my @a_docids = `$lucene_searcher --parser_cfg=$parser_cfg --store_dir=$store_dir --list --query=$docid_field`;
    chomp @a_docids;
    
    
    for (my $i = 0; $i <= $#a_docids; $i = $i + $SLICE_LUCENE_QUERY)
    {
		my $last = $i + $SLICE_LUCENE_QUERY - 1;
		$last = $#a_docids if ($last > $#a_docids);
		
		my @a_slice = @a_docids[$i..$last]; 
		my $lucene_query = "$docid_field:" . join (" $docid_field:" , @a_slice);

    	
		my $cde_get_definitions = "$lucene_searcher -query '$lucene_query' -store $store_dir -parser $parser_cfg --output_fields=id,$docid_field,length,product,go,ec";
        print STDERR "CMD - $cde_get_definitions\n" if ($o_param->IsDefined('verbose'));
        my $raw_results = `$cde_get_definitions`;
        
        
        my @a_blocks = split (/\/\/\n/, $raw_results);
        foreach my $block (@a_blocks)
        {
			my @a_results = split (/\n/,$block);
        
        my $test_match = @a_results;
        my ($accession,$id,$product,$gos,$length,$ecs) = ('','','','',0,'');
		my $ec_numbers = '';
		my ($go_molfunc,$go_biolproc,$go_cellcomp) = ('','','');
		
		if ($test_match > 1)
		{
			foreach my $result (@a_results)
			{
				chomp $result;
				next unless ($result =~ /=/);
				
				#print STDERR $result;
				#exit;
				($accession) = ($result =~ /^$docid_field=([\w\.]*)/) if ($result =~ /^$docid_field=/);
				
				if (!defined $h_analyses{$accession})
				{
					$h_analyses{$accession} = {};
					 print STDERR "INFO - Analyse $accession\n" if ($o_param->IsDefined('verbose'));
				}
				($id)        = ($result =~ /^id=(\S*)/) if ($result =~ /^id=/);
				($product)   = ($result =~ /^product=(.*)/) if ($result =~ /^product=/);
				($gos)       = ($result =~ /^go=(.*)/) if ($result =~ /^go=/);
				($ecs)       = ($result =~ /^ec=(.*)/) if ($result =~ /^ec=/);
				($length)    = ($result =~ /^length=(\d*)/) if ($result =~ /^length=/);
	            $product =~ s/\/\//;/g;
				$gos      =~ s/\/\//;/g;
	            $ecs      =~ s/\/\//;/g;
	            $product =~ s/\/\//;/g;

	            my @a_gos = split (/IPR\d+;/, $gos);
	            foreach my $raw_go_term (@a_gos)
	            {
					my @a_raw_go_terms = split ('\),', $raw_go_term);
					foreach my $go_term (@a_raw_go_terms)
					{
	                
	                $go_term =~ s/\s*;\s*$//;
					$go_term =~ s/^\s+//g;
					$go_term .= ')' if ($go_term !~ /\)$/);
	                

					
					if ($go_term =~ /Molecular Function:\s+(.+)/)
					{
						my $desc = $1;
						my ($go_number) = ($desc =~ /(GO:\d+)/);
						next if ($go_molfunc =~ /$go_number/);
						$go_molfunc .= $desc . ';';
					}
					elsif ($go_term =~ /Biological Process:\s+(.+)/)
					{
						my $desc = $1;
						my ($go_number) = ($desc =~ /(GO:\d+)/);
						next if ($go_biolproc =~ /$go_number/);
						
						$go_biolproc .= $desc . ';';
					}
					elsif ($go_term =~ /Cellular Component:\s+(.+)/)
					{
						my $desc = $1;
						my ($go_number) = ($desc =~ /(GO:\d+)/);
						next if ($go_cellcomp =~ /$go_number/);
						
						$go_cellcomp .= $desc . ';';
					}
				}
				}
	        
				my @a_ecs = split (/;/, $ecs);
	            foreach my $ec_term (@a_ecs)
	            {
		                $ec_term =~ s/\s*;\s*$//;
		                $ec_numbers .= $ec_term . ';';
	            }
	            
			}
			$h_deflines{$accession} = "$accession\t$id\t$length\t$product\t$ec_numbers\t$go_molfunc\t$go_biolproc\t$go_cellcomp\t";
			
		 }
		}

         my $cde_get_xml = "$lucene_searcher -no_output_field_name -query '$lucene_query' -store $store_dir -parser $parser_cfg";
         my $cde_convert2xls = $o_param->Get('xsltproc') . ' ' . $o_param->Get('lucene2tab_xsl') . ' - ';
        
		 print STDERR "CMD - $cde_get_xml | $cde_convert2xls\n" if ($o_param->IsDefined('verbose')); 
         my @a_lines = `$cde_get_xml | $cde_convert2xls`;
         chomp @a_lines;

         my $header_line = shift @a_lines;
         my $accession;
         foreach my $line (@a_lines)
         {
                my @a_fields = split("\t", $line);
                my $dbname = $a_fields[24];
                my $score = $a_fields[13];
                next if (! defined $dbname);
                
                my %h_query_ids = (id => $a_fields[0], accession => $a_fields[1], locus_tag => $a_fields[2]);
                $accession = $h_query_ids{$docid_field};
                #$line =~ s/^$a_fields[0]\t$a_fields[1]\t$a_fields[2]\t//;
                                
                #print "$dbname\n";
				$dbname =~ s/^.+\///;
				$line =~ s/\S+\/$dbname/$dbname/;
                next if (! grep (/^$dbname$/,@a_dbnames));
                $h_dbnames{$dbname} = $header_line if (!exists $h_dbnames{$dbname});
                $h_analyses{$accession}->{$dbname} = $line if (!exists $h_analyses{$accession}->{$dbname});
                
                if (!exists $h_absolutebesthit{$accession} && defined $score && $score ne '')
                {
					$h_absolutebesthit{$accession} = {line => $line, score => $score} ;
				}
				elsif (exists $h_absolutebesthit{$accession} && ($score > $h_absolutebesthit{$accession}->{score}))
				{
						$h_absolutebesthit{$accession} = {line => $line, score => $score} ;
				}
                
         }
         
         print STDERR "INFO - Best Hit - " . $h_absolutebesthit{$accession}->{line} ."\n" if ($o_param->IsDefined('verbose'));
         
         
         
		 
		 
	#	 last if (scalar (keys %h_deflines) > 3);
		 
	}
	my $outdir = $o_param->Get('outdir');
    mkdir $outdir unless (-d $outdir);
    my $date = `date +%Y%m%d`;
    chomp $date;
    my $type = basename($store_dir);
    my $outfile = "$outdir/$date.$type.blast.besthits.xls";
    my $fh_out = &GetStreamOut($outfile);

    print STDERR "INFO - Print output file $outfile\n" if ($o_param->IsDefined('verbose'));
  	print $fh_out "$docid_field\tId\tLength\tProduct\tEC Numbers\tGO Molecular function\tGO Biological Process\tGO Cellular Component\t";
    my $absolutebestof_numfields = 0;
    foreach my $dbname (sort (@a_dbnames))
    {
        my $header = $h_dbnames{$dbname} if (defined $h_dbnames{$dbname});
        $header =~ s/^\s+//g;
        if ($header ne '')
        {
		
        my @a_fields = split("\t", $header);
        print $fh_out "$dbname\t" x scalar @a_fields;
        $absolutebestof_numfields = scalar @a_fields;
		}
    }
    print $fh_out "BEST HIT\t" x $absolutebestof_numfields;
    
    print $fh_out "\n";
    
    print $fh_out "\n";
    print $fh_out "\t" x 8;
    my $absolutebestof_header = '';
    foreach my $dbname (sort (@a_dbnames))
    {
         my $header = $h_dbnames{$dbname} if (defined $h_dbnames{$dbname});
         if ($header ne '')
         {
			$header =~ s/^\s+//g;
			$absolutebestof_header = $header;
			print $fh_out "$header\t";
		}
		else
		{
				print STDERR "INFO - No hit for $dbname\n" if ($o_param->IsDefined('verbose'));
		}
    }
    print $fh_out $absolutebestof_header;
    print $fh_out "\n";

	foreach my $accession (keys %h_deflines)
	{
		print $fh_out $h_deflines{$accession} ;
		foreach my $dbname (sort (@a_dbnames))
         {
             my $header   = $h_dbnames{$dbname};
             $header =~ s/^\s+//g;
            if ($header ne '') #Il n'y a aucun hit pour aucune des accessions contre cette banque
			{
             my @a_fields = split("\t", $header);
             
             my $line     = "\t" x scalar @a_fields;
             
             
             if (defined $h_analyses{$accession}->{$dbname} && $h_analyses{$accession}->{$dbname} ne '')
             {
				$line = $h_analyses{$accession}->{$dbname} ;
			 }
			 
			 $line .= "\t" if ($line !~ /\t$/);
			 print $fh_out "$line";
            } 

         }
         
         if (defined $h_absolutebesthit{$accession})
         {
				print $fh_out $h_absolutebesthit{$accession}->{line};
		 }
         print $fh_out "\n";
         
      }

    $fh_out->close;
    
    return;
}

sub Usage
{
    print STDOUT <<END;

Usage:   $0  --outdir=<existing output directory> --store=<lucene store> 

    --light_entry_xml : delete entry_features element
    --parser_cfg : Lucene configuration file
    --store_dir : Lucene storage directory
    --outdir : output directory
    
    --output_field : dumped fields
    --outfile_suffix : add this suffix to new files
    
    --table : print a summary table (id,length, product, go terms, best hits)
    --faa	: dump proteins in multifasta format
    
END
}
